@@ -19,7 +19,6 @@
  *
  * Written by: Ray Strode <rstrode@redhat.com>
  */
-#include "config.h"
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -55,6 +54,7 @@
 #include "ply-trigger.h"
 #include "ply-utils.h"
 #include "ply-progress.h"
+#include "ply-kmsg-reader.h"
 
 #define BOOT_DURATION_FILE     PLYMOUTH_TIME_DIRECTORY "/boot-duration"
 #define SHUTDOWN_DURATION_FILE PLYMOUTH_TIME_DIRECTORY "/shutdown-duration"
@@ -79,6 +79,7 @@ typedef struct
         ply_event_loop_t       *loop;
         ply_boot_server_t      *boot_server;
         ply_boot_splash_t      *boot_splash;
+        ply_kmsg_reader_t      *kmsg_reader;
         ply_terminal_session_t *session;
         ply_buffer_t           *boot_buffer;
         ply_progress_t         *progress;
@@ -93,6 +94,7 @@ typedef struct
 
         ply_trigger_t          *deactivate_trigger;
         ply_trigger_t          *quit_trigger;
+        ply_trigger_t          *kmsg_trigger;
 
         double                  start_time;
         double                  splash_delay;
@@ -108,6 +110,7 @@ typedef struct
         uint32_t                is_inactive : 1;
         uint32_t                is_shown : 1;
         uint32_t                should_force_details : 1;
+        uint32_t                should_force_default_splash : 1;
         uint32_t                splash_is_becoming_idle : 1;
 
         char                   *override_splash_path;
@@ -138,6 +141,7 @@ static void on_error_message (ply_buffer_t *debug_buffer,
                               size_t        number_of_bytes);
 static ply_buffer_t *debug_buffer;
 static char *debug_buffer_path = NULL;
+static char *boot_log_file = NULL;
 static char *pid_file = NULL;
 static void toggle_between_splash_and_details (state_t *state);
 #ifdef PLY_ENABLE_SYSTEMD_INTEGRATION
@@ -155,6 +159,8 @@ static void on_backspace (state_t *state);
 static void on_quit (state_t       *state,
                      bool           retain_splash,
                      ply_trigger_t *quit_trigger);
+static void on_new_kmsg_message (state_t        *state,
+                                 kmsg_message_t *kmsg_message);
 static bool sh_is_init (state_t *state);
 static void cancel_pending_delayed_show (state_t *state);
 static void prepare_logging (state_t *state);
@@ -206,6 +212,8 @@ on_change_mode (state_t    *state,
                 state->mode = PLY_BOOT_SPLASH_MODE_SYSTEM_UPGRADE;
         else if (strcmp (mode, "firmware-upgrade") == 0)
                 state->mode = PLY_BOOT_SPLASH_MODE_FIRMWARE_UPGRADE;
+        else if (strcmp (mode, "system-reset") == 0)
+                state->mode = PLY_BOOT_SPLASH_MODE_SYSTEM_RESET;
         else
                 return;
 
@@ -262,9 +270,9 @@ show_messages (state_t *state)
 }
 
 static bool
-get_theme_path (const char  *splash_string,
-                const char  *configured_theme_dir,
-                char       **theme_path)
+get_theme_path (const char *splash_string,
+                const char *configured_theme_dir,
+                char      **theme_path)
 {
         const char *paths[] = { PLYMOUTH_RUNTIME_THEME_PATH,
                                 configured_theme_dir,
@@ -311,18 +319,18 @@ load_settings (state_t    *state,
         if (splash_string != NULL) {
                 char *configured_theme_dir;
                 configured_theme_dir = ply_key_file_get_value (key_file, "Daemon",
-                                                                         "ThemeDir");
+                                                               "ThemeDir");
                 get_theme_path (splash_string, configured_theme_dir, theme_path);
                 free (configured_theme_dir);
         }
 
         if (isnan (state->splash_delay)) {
-                state->splash_delay = ply_key_file_get_double(key_file, "Daemon", "ShowDelay", NAN);
+                state->splash_delay = ply_key_file_get_double (key_file, "Daemon", "ShowDelay", NAN);
                 ply_trace ("Splash delay is set to %lf", state->splash_delay);
         }
 
         if (isnan (state->device_timeout)) {
-                state->device_timeout = ply_key_file_get_double(key_file, "Daemon", "DeviceTimeout", NAN);
+                state->device_timeout = ply_key_file_get_double (key_file, "Daemon", "DeviceTimeout", NAN);
                 ply_trace ("Device timeout is set to %lf", state->device_timeout);
         }
 
@@ -590,7 +598,7 @@ on_hide_message (state_t    *state,
                         free (list_message);
                         ply_list_remove_node (state->messages, node);
                         if (state->boot_splash != NULL) {
-                            ply_boot_splash_hide_message (state->boot_splash, message);
+                                ply_boot_splash_hide_message (state->boot_splash, message);
                         }
                 }
                 node = next_node;
@@ -666,7 +674,7 @@ on_newroot (state_t    *state,
         chroot (".");
         chdir ("/");
         /* Update local now that we have /usr/share/locale available */
-        setlocale(LC_ALL, "");
+        setlocale (LC_ALL, "");
         ply_progress_load_cache (state->progress, get_cache_file_for_mode (state->mode));
         if (state->boot_splash != NULL)
                 ply_boot_splash_root_mounted (state->boot_splash);
@@ -688,6 +696,7 @@ get_cache_file_for_mode (ply_boot_splash_mode_t mode)
         case PLY_BOOT_SPLASH_MODE_UPDATES:
         case PLY_BOOT_SPLASH_MODE_SYSTEM_UPGRADE:
         case PLY_BOOT_SPLASH_MODE_FIRMWARE_UPGRADE:
+        case PLY_BOOT_SPLASH_MODE_SYSTEM_RESET:
                 filename = NULL;
                 break;
         case PLY_BOOT_SPLASH_MODE_INVALID:
@@ -708,16 +717,21 @@ get_log_file_for_state (state_t *state)
 
         switch (state->mode) {
         case PLY_BOOT_SPLASH_MODE_BOOT_UP:
-                if (state->no_boot_log)
+                if (state->no_boot_log) {
                         filename = NULL;
-                else
-                        filename = PLYMOUTH_LOG_DIRECTORY "/boot.log";
+                } else {
+                        if (boot_log_file == NULL)
+                                filename = PLYMOUTH_LOG_DIRECTORY "/boot.log";
+                        else
+                                filename = boot_log_file;
+                }
                 break;
         case PLY_BOOT_SPLASH_MODE_SHUTDOWN:
         case PLY_BOOT_SPLASH_MODE_REBOOT:
         case PLY_BOOT_SPLASH_MODE_UPDATES:
         case PLY_BOOT_SPLASH_MODE_SYSTEM_UPGRADE:
         case PLY_BOOT_SPLASH_MODE_FIRMWARE_UPGRADE:
+        case PLY_BOOT_SPLASH_MODE_SYSTEM_RESET:
                 filename = _PATH_DEVNULL;
                 break;
         case PLY_BOOT_SPLASH_MODE_INVALID:
@@ -745,6 +759,7 @@ get_log_spool_file_for_mode (ply_boot_splash_mode_t mode)
         case PLY_BOOT_SPLASH_MODE_UPDATES:
         case PLY_BOOT_SPLASH_MODE_SYSTEM_UPGRADE:
         case PLY_BOOT_SPLASH_MODE_FIRMWARE_UPGRADE:
+        case PLY_BOOT_SPLASH_MODE_SYSTEM_RESET:
                 filename = NULL;
                 break;
         case PLY_BOOT_SPLASH_MODE_INVALID:
@@ -858,7 +873,7 @@ sh_is_init (state_t *state)
         if (init_string) {
                 length = strlen (init_string);
                 if (length > 2 && init_string[length - 2] == 's' &&
-                                  init_string[length - 1] == 'h')
+                    init_string[length - 1] == 'h')
                         result = true;
 
                 free (init_string);
@@ -907,10 +922,54 @@ plymouth_should_show_default_splash (state_t *state)
                 return true;
         }
 
+        if (state->should_force_default_splash) {
+                ply_trace ("using default splash because kernel command line has option \"plymouth.graphical\"");
+                return true;
+        }
+
         ply_trace ("no default splash because kernel command line lacks \"splash\" or \"rhgb\"");
         return false;
 }
 
+static void
+on_reload (state_t *state)
+{
+        ply_trace ("reloading");
+        if (state->boot_splash != NULL) {
+                ply_boot_splash_hide (state->boot_splash);
+                ply_boot_splash_free (state->boot_splash);
+                state->boot_splash = NULL;
+        }
+
+        free (state->override_splash_path);
+        state->override_splash_path = NULL;
+        free (state->system_default_splash_path);
+        state->system_default_splash_path = NULL;
+        free (state->distribution_default_splash_path);
+        state->distribution_default_splash_path = NULL;
+
+        find_override_splash (state);
+        find_system_default_splash (state);
+        find_distribution_default_splash (state);
+
+        if (state->is_inactive) {
+                ply_trace ("reload while inactive");
+                return;
+        }
+
+        if (!state->is_shown) {
+                ply_trace ("reload while not shown");
+                return;
+        }
+
+        if (state->showing_details) {
+                show_detailed_splash (state);
+        } else {
+                show_default_splash (state);
+        }
+}
+
+
 static void
 on_show_splash (state_t *state)
 {
@@ -939,6 +998,14 @@ on_show_splash (state_t *state)
         if (!state->is_attached && state->should_be_attached && has_displays)
                 attach_to_running_session (state);
 
+        if (state->local_console_terminal != NULL)
+                ply_terminal_set_mode (state->local_console_terminal, PLY_TERMINAL_MODE_GRAPHICS);
+
+#ifdef PLY_ENABLE_SYSTEMD_INTEGRATION
+        if (state->is_attached)
+                tell_systemd_to_print_details (state);
+#endif
+
         if (has_displays) {
                 ply_trace ("at least one display already available, so loading splash");
                 show_splash (state);
@@ -1018,25 +1085,25 @@ static void
 on_keyboard_removed (state_t        *state,
                      ply_keyboard_t *keyboard)
 {
-    ply_trace ("no longer listening for keystrokes");
-    ply_keyboard_remove_input_handler (keyboard,
-                                       (ply_keyboard_input_handler_t)
-                                       on_keyboard_input);
-    ply_trace ("no longer listening for escape");
-    ply_keyboard_remove_escape_handler (keyboard,
-                                        (ply_keyboard_escape_handler_t)
-                                        on_escape_pressed);
-    ply_trace ("no longer listening for backspace");
-    ply_keyboard_remove_backspace_handler (keyboard,
-                                           (ply_keyboard_backspace_handler_t)
-                                           on_backspace);
-    ply_trace ("no longer listening for enter");
-    ply_keyboard_remove_enter_handler (keyboard,
-                                       (ply_keyboard_enter_handler_t)
-                                       on_enter);
+        ply_trace ("no longer listening for keystrokes");
+        ply_keyboard_remove_input_handler (keyboard,
+                                           (ply_keyboard_input_handler_t)
+                                           on_keyboard_input);
+        ply_trace ("no longer listening for escape");
+        ply_keyboard_remove_escape_handler (keyboard,
+                                            (ply_keyboard_escape_handler_t)
+                                            on_escape_pressed);
+        ply_trace ("no longer listening for backspace");
+        ply_keyboard_remove_backspace_handler (keyboard,
+                                               (ply_keyboard_backspace_handler_t)
+                                               on_backspace);
+        ply_trace ("no longer listening for enter");
+        ply_keyboard_remove_enter_handler (keyboard,
+                                           (ply_keyboard_enter_handler_t)
+                                           on_enter);
 
-    if (state->boot_splash != NULL)
-            ply_boot_splash_unset_keyboard (state->boot_splash);
+        if (state->boot_splash != NULL)
+                ply_boot_splash_unset_keyboard (state->boot_splash);
 }
 
 static void
@@ -1159,8 +1226,10 @@ hide_splash (state_t *state)
 
         ply_boot_splash_hide (state->boot_splash);
 
-        if (state->local_console_terminal != NULL)
+        if (state->local_console_terminal != NULL) {
                 ply_terminal_set_mode (state->local_console_terminal, PLY_TERMINAL_MODE_TEXT);
+                ply_terminal_set_buffered_input (state->local_console_terminal);
+        }
 }
 
 static void
@@ -1227,7 +1296,6 @@ deactivate_console (state_t *state)
         /* do not let any tty opened where we could write after deactivate */
         if (ply_kernel_command_line_has_argument ("plymouth.debug"))
                 ply_logger_close_file (ply_logger_get_error_default ());
-
 }
 
 static void
@@ -1368,10 +1436,6 @@ on_quit (state_t       *state,
         state->quit_trigger = quit_trigger;
         state->should_retain_splash = retain_splash;
 
-#ifdef PLY_ENABLE_SYSTEMD_INTEGRATION
-        tell_systemd_to_stop_printing_details (state);
-#endif
-
         ply_trace ("closing log");
         if (state->session != NULL)
                 ply_terminal_session_close_log (state->session);
@@ -1397,6 +1461,18 @@ on_quit (state_t       *state,
         }
 }
 
+void
+on_new_kmsg_message (state_t        *state,
+                     kmsg_message_t *kmsg_message)
+{
+        ply_buffer_append (state->boot_buffer, "%s\n", kmsg_message->message);
+
+        if (state->boot_splash != NULL) {
+                ply_boot_splash_update_output (state->boot_splash, kmsg_message->message, strlen (kmsg_message->message));
+                ply_boot_splash_update_output (state->boot_splash, "\n", 1);
+        }
+}
+
 static bool
 on_has_active_vt (state_t *state)
 {
@@ -1431,6 +1507,7 @@ start_boot_server (state_t *state)
                                       (ply_boot_server_reactivate_handler_t) on_reactivate,
                                       (ply_boot_server_quit_handler_t) on_quit,
                                       (ply_boot_server_has_active_vt_handler_t) on_has_active_vt,
+                                      (ply_boot_server_reload_handler_t) on_reload,
                                       state);
 
         if (!ply_boot_server_listen (server)) {
@@ -1445,6 +1522,19 @@ start_boot_server (state_t *state)
         return server;
 }
 
+static bool
+validate_input (state_t    *state,
+                const char *entry_text,
+                const char *add_text)
+{
+        bool input_valid;
+
+        if (!state->boot_splash)
+                return true;
+        input_valid = ply_boot_splash_validate_input (state->boot_splash, entry_text, add_text);
+        return input_valid;
+}
+
 
 static void
 update_display (state_t *state)
@@ -1462,10 +1552,18 @@ update_display (state_t *state)
                         ply_boot_splash_display_password (state->boot_splash,
                                                           entry_trigger->prompt,
                                                           bullets);
+                        ply_boot_splash_display_prompt (state->boot_splash,
+                                                        entry_trigger->prompt,
+                                                        ply_buffer_get_bytes (state->entry_buffer),
+                                                        true);
                 } else if (entry_trigger->type == PLY_ENTRY_TRIGGER_TYPE_QUESTION) {
                         ply_boot_splash_display_question (state->boot_splash,
                                                           entry_trigger->prompt,
                                                           ply_buffer_get_bytes (state->entry_buffer));
+                        ply_boot_splash_display_prompt (state->boot_splash,
+                                                        entry_trigger->prompt,
+                                                        ply_buffer_get_bytes (state->entry_buffer),
+                                                        false);
                 } else {
                         ply_trace ("unkown entry type");
                 }
@@ -1498,7 +1596,17 @@ static void
 on_escape_pressed (state_t *state)
 {
         ply_trace ("escape key pressed");
-        toggle_between_splash_and_details (state);
+        bool has_vt_consoles = true;
+
+        if (state->local_console_terminal != NULL) {
+                if (!ply_terminal_is_vt (state->local_console_terminal))
+                        has_vt_consoles = false;
+        } else {
+                has_vt_consoles = false;
+        }
+
+        if (validate_input (state, "", "\e") && has_vt_consoles == true)
+                toggle_between_splash_and_details (state);
 }
 
 static void
@@ -1510,12 +1618,16 @@ on_keyboard_input (state_t    *state,
 
         node = ply_list_get_first_node (state->entry_triggers);
         if (node) { /* \x3 (ETX) is Ctrl+C and \x4 (EOT) is Ctrl+D */
+                if (!validate_input (state, ply_buffer_get_bytes (state->entry_buffer), keyboard_input))
+                        return;
                 if (character_size == 1 && (keyboard_input[0] == '\x3' || keyboard_input[0] == '\x4')) {
                         ply_entry_trigger_t *entry_trigger = ply_list_node_get_data (node);
                         ply_trigger_pull (entry_trigger->trigger, "\x3");
                         ply_buffer_clear (state->entry_buffer);
                         ply_list_remove_node (state->entry_triggers, node);
                         free (entry_trigger);
+                } else if (character_size >= 2 && keyboard_input[0] == '\033') {
+                        /* Ignore escape sequences */
                 } else {
                         ply_buffer_append_bytes (state->entry_buffer, keyboard_input, character_size);
                 }
@@ -1538,28 +1650,17 @@ on_keyboard_input (state_t    *state,
 static void
 on_backspace (state_t *state)
 {
-        ssize_t bytes_to_remove;
-        ssize_t previous_character_size;
-        const char *bytes;
+        char *bytes;
         size_t size;
+        size_t capacity;
         ply_list_node_t *node = ply_list_get_first_node (state->entry_triggers);
 
         if (!node) return;
 
-        bytes = ply_buffer_get_bytes (state->entry_buffer);
-        size = ply_buffer_get_size (state->entry_buffer);
-        if (size == 0)
-                return;
-
-        bytes_to_remove = MIN (size, PLY_UTF8_CHARACTER_SIZE_MAX);
-        while ((previous_character_size = ply_utf8_character_get_size (bytes + size - bytes_to_remove, bytes_to_remove)) < bytes_to_remove) {
-                if (previous_character_size > 0)
-                        bytes_to_remove -= previous_character_size;
-                else
-                        bytes_to_remove--;
+        ply_buffer_borrow_bytes (state->entry_buffer, &bytes, &size, &capacity) {
+                ply_utf8_string_remove_last_character (&bytes, &size);
         }
 
-        ply_buffer_remove_bytes_at_end (state->entry_buffer, bytes_to_remove);
         update_display (state);
 }
 
@@ -1573,6 +1674,8 @@ on_enter (state_t    *state,
         if (node) {
                 ply_entry_trigger_t *entry_trigger = ply_list_node_get_data (node);
                 const char *reply_text = ply_buffer_get_bytes (state->entry_buffer);
+                if (!validate_input (state, reply_text, "\n"))
+                        return;
                 ply_trigger_pull (entry_trigger->trigger, reply_text);
                 ply_buffer_clear (state->entry_buffer);
                 ply_list_remove_node (state->entry_triggers, node);
@@ -1796,6 +1899,18 @@ attach_to_running_session (state_t *state)
                 return false;
         }
 
+        if (state->kmsg_reader == NULL) {
+                ply_trace ("Creating new kmsg reader");
+                state->kmsg_reader = ply_kmsg_reader_new ();
+
+                ply_kmsg_reader_watch_for_messages (state->kmsg_reader,
+                                                    (ply_kmsg_reader_message_handler_t)
+                                                    on_new_kmsg_message,
+                                                    state);
+        }
+
+        ply_kmsg_reader_start (state->kmsg_reader);
+
 #ifdef PLY_ENABLE_SYSTEMD_INTEGRATION
         tell_systemd_to_print_details (state);
 #endif
@@ -1820,6 +1935,9 @@ detach_from_running_session (state_t *state)
         tell_systemd_to_stop_printing_details (state);
 #endif
 
+        ply_trace ("stopping kmsg reader");
+        ply_kmsg_reader_stop (state->kmsg_reader);
+
         ply_trace ("detaching from terminal session");
         ply_terminal_session_detach (state->session);
         state->is_redirected = false;
@@ -1906,6 +2024,9 @@ check_logging (state_t *state)
 
         ply_trace ("checking if console messages should be redirected and logged");
 
+        if (!boot_log_file)
+                boot_log_file = ply_kernel_command_line_get_key_value ("plymouth.boot-log=");
+
         kernel_no_log = ply_kernel_command_line_has_argument ("plymouth.nolog");
         if (kernel_no_log)
                 state->no_boot_log = true;
@@ -1972,8 +2093,10 @@ initialize_environment (state_t *state)
 
                 ply_trace ("checking if '%s' exists", state->default_tty);
                 if (!ply_character_device_exists (state->default_tty)) {
-                        ply_trace ("nope, forcing details mode");
-                        state->should_force_details = true;
+                        if (!state->should_force_default_splash) {
+                                ply_trace ("nope, forcing details mode");
+                                state->should_force_details = true;
+                        }
 
                         state->default_tty = find_fallback_tty (state);
                         ply_trace ("going to go with '%s'", state->default_tty);
@@ -2042,7 +2165,7 @@ on_crash (int signum)
 
         ioctl (fd, KDSETMODE, KD_TEXT);
 
-        write (fd, show_cursor_sequence, sizeof (show_cursor_sequence) - 1);
+        write (fd, show_cursor_sequence, sizeof(show_cursor_sequence) - 1);
 
         tcgetattr (fd, &term_attributes);
 
@@ -2131,6 +2254,7 @@ main (int    argc,
         bool no_daemon = false;
         bool debug = false;
         bool ignore_serial_consoles = false;
+        bool graphical_boot = false;
         bool attach_to_session;
         ply_daemon_handle_t *daemon_handle = NULL;
         char *mode_string = NULL;
@@ -2144,8 +2268,8 @@ main (int    argc,
         state.loop = ply_event_loop_get_default ();
 
         /* Initialize the translations if they are available (!initrd) */
-        if (ply_directory_exists (PLYMOUTH_LOCALE_DIRECTORY))
-                setlocale(LC_ALL, "");
+        if (ply_file_exists (PLYMOUTH_LOCALE_DIRECTORY "/nl/LC_MESSAGES/plymouth.mo"))
+                setlocale (LC_ALL, "");
 
         ply_command_parser_add_options (state.command_parser,
                                         "help", "This help message", PLY_COMMAND_OPTION_TYPE_FLAG,
@@ -2159,6 +2283,7 @@ main (int    argc,
                                         "tty", "TTY to use instead of default", PLY_COMMAND_OPTION_TYPE_STRING,
                                         "no-boot-log", "Do not write boot log file", PLY_COMMAND_OPTION_TYPE_FLAG,
                                         "ignore-serial-consoles", "Ignore serial consoles", PLY_COMMAND_OPTION_TYPE_FLAG,
+                                        "graphical-boot", "Use graphical splashes even if the kernel console is not a VT", PLY_COMMAND_OPTION_TYPE_FLAG,
                                         NULL);
 
         if (!ply_command_parser_parse_arguments (state.command_parser, state.loop, argv, argc)) {
@@ -2180,7 +2305,9 @@ main (int    argc,
                                         "no-daemon", &no_daemon,
                                         "debug", &debug,
                                         "ignore-serial-consoles", &ignore_serial_consoles,
+                                        "graphical-boot", &graphical_boot,
                                         "debug-file", &debug_buffer_path,
+                                        "boot-log", &boot_log_file,
                                         "pid-file", &pid_file,
                                         "tty", &tty,
                                         "kernel-command-line", &kernel_command_line,
@@ -2214,6 +2341,8 @@ main (int    argc,
                         state.mode = PLY_BOOT_SPLASH_MODE_SYSTEM_UPGRADE;
                 else if (strcmp (mode_string, "firmware-upgrade") == 0)
                         state.mode = PLY_BOOT_SPLASH_MODE_FIRMWARE_UPGRADE;
+                else if (strcmp (mode_string, "system-reset") == 0)
+                        state.mode = PLY_BOOT_SPLASH_MODE_SYSTEM_RESET;
                 else
                         state.mode = PLY_BOOT_SPLASH_MODE_BOOT_UP;
 
@@ -2251,6 +2380,11 @@ main (int    argc,
         signal (SIGABRT, on_crash);
         signal (SIGSEGV, on_crash);
 
+        if (graphical_boot || ply_kernel_command_line_has_argument ("plymouth.graphical")) {
+                state.should_force_default_splash = true;
+                ignore_serial_consoles = true;
+        }
+
         /* before do anything we need to make sure we have a working
          * environment.
          */
@@ -2331,6 +2465,11 @@ main (int    argc,
             (getenv ("DISPLAY") != NULL))
                 device_manager_flags |= PLY_DEVICE_MANAGER_FLAGS_IGNORE_UDEV;
 
+        if ((ply_kernel_command_line_has_argument ("plymouth.force-frame-buffer-on-boot")) &&
+            state.mode != PLY_BOOT_SPLASH_MODE_SHUTDOWN &&
+            state.mode != PLY_BOOT_SPLASH_MODE_REBOOT)
+                device_manager_flags |= PLY_DEVICE_MANAGER_FLAGS_FORCE_FRAME_BUFFER;
+
         if (!plymouth_should_show_default_splash (&state)) {
                 /* don't bother listening for udev events or setting up a graphical renderer
                  * if we're forcing details */
@@ -2378,4 +2517,3 @@ main (int    argc,
 
         return exit_code;
 }
-/* vim: set ts=4 ts=4 sw=4 expandtab autoindent cindent cino={.5s,(0: */
